from typing import Optional, Union

import zarrita.codecs  # noqa: F401
from zarrita.array import (  # noqa: F401
    Array,
    ArrayRuntimeConfiguration,
    runtime_configuration,
)
from zarrita.array_v2 import ArrayV2  # noqa: F401
from zarrita.group import Group  # noqa: F401
from zarrita.group_v2 import GroupV2  # noqa: F401
from zarrita.store import (  # noqa: F401
    LocalStore,
    RemoteStore,
    Store,
    StoreLike,
    StorePath,
    make_store_path,
)
from zarrita.sync import sync as _sync


async def open_auto_async(
    store: StoreLike,
    runtime_configuration_: Optional[ArrayRuntimeConfiguration] = None,
) -> Union[Array, ArrayV2, Group, GroupV2]:
    store_path = make_store_path(store)
    try:
        return await Group.open_or_array(
            store_path, runtime_configuration=runtime_configuration_
        )
    except KeyError:
        return await GroupV2.open_or_array(store_path)


def open_auto(
    store: StoreLike,
    runtime_configuration_: Optional[ArrayRuntimeConfiguration] = None,
) -> Union[Array, ArrayV2, Group, GroupV2]:
    return _sync(open_auto_async(store, runtime_configuration_))



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Awaitable, Callable, List, Literal, Tuple, Union

import numpy as np
from attr import asdict, frozen
from numcodecs.blosc import Blosc
from numcodecs.gzip import GZip

from zarrita.common import BytesLike, to_thread
from zarrita.metadata import (
    BloscCodecConfigurationMetadata,
    BloscCodecMetadata,
    CodecMetadata,
    EndianCodecConfigurationMetadata,
    EndianCodecMetadata,
    GzipCodecConfigurationMetadata,
    GzipCodecMetadata,
    ShardingCodecConfigurationMetadata,
    ShardingCodecMetadata,
    TransposeCodecConfigurationMetadata,
    TransposeCodecMetadata,
    data_type_to_numpy,
)
from zarrita.value_handle import (
    ArrayValueHandle,
    BufferValueHandle,
    NoneValueHandle,
    ValueHandle,
)

if TYPE_CHECKING:
    from zarrita.array import CoreArrayMetadata


async def _needs_array(
    chunk_value_handle: ValueHandle,
    array_metadata: "CoreArrayMetadata",
    f: Callable[
        [np.ndarray, "CoreArrayMetadata"], Awaitable[Union[None, np.ndarray, BytesLike]]
    ],
):
    chunk_array = await chunk_value_handle.toarray()
    if chunk_array is None:
        return NoneValueHandle()
    if chunk_array.dtype.name != array_metadata.dtype.name:
        chunk_array = chunk_array.view(dtype=array_metadata.dtype)
    result = await f(chunk_array, array_metadata)
    if result is None:
        return NoneValueHandle()
    elif (
        isinstance(result, bytes)
        or isinstance(result, bytearray)
        or isinstance(result, memoryview)
    ):
        return BufferValueHandle(result)
    elif isinstance(result, np.ndarray):
        return ArrayValueHandle(result)


async def _needs_bytes(
    chunk_value_handle: ValueHandle,
    array_metadata: "CoreArrayMetadata",
    f: Callable[
        [BytesLike, "CoreArrayMetadata"], Awaitable[Union[None, np.ndarray, BytesLike]]
    ],
):
    chunk_bytes = await chunk_value_handle.tobytes()
    if chunk_bytes is None:
        return NoneValueHandle()
    result = await f(chunk_bytes, array_metadata)
    if result is None:
        return NoneValueHandle()
    elif (
        isinstance(result, bytes)
        or isinstance(result, bytearray)
        or isinstance(result, memoryview)
    ):
        return BufferValueHandle(result)
    elif isinstance(result, np.ndarray):
        return ArrayValueHandle(result)


class Codec(ABC):
    supports_partial_decode: bool
    supports_partial_encode: bool
    input_type: Literal["bytes", "array"]
    output_type: Literal["bytes", "array"]

    @abstractmethod
    async def decode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        pass

    @abstractmethod
    async def encode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        pass

    @staticmethod
    def codecs_from_metadata(codecs_metadata: List["CodecMetadata"]) -> List["Codec"]:
        out: List[Codec] = []
        for codec_metadata in codecs_metadata:
            if codec_metadata.name == "blosc":
                out.append(BloscCodec.from_metadata(codec_metadata))
            elif codec_metadata.name == "gzip":
                out.append(GzipCodec.from_metadata(codec_metadata))
            elif codec_metadata.name == "transpose":
                out.append(TransposeCodec.from_metadata(codec_metadata))
            elif codec_metadata.name == "endian":
                out.append(EndianCodec.from_metadata(codec_metadata))
            elif codec_metadata.name == "sharding_indexed":
                from zarrita.sharding import ShardingCodec

                out.append(ShardingCodec.from_metadata(codec_metadata))
            else:
                raise RuntimeError(f"Unsupported codec: {codec_metadata}")
        return out


class ArrayArrayCodec(Codec):
    input_type = "array"
    output_type = "array"

    async def decode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return await _needs_array(chunk_value_handle, array_metadata, self.inner_decode)

    @abstractmethod
    async def inner_decode(
        self,
        chunk_array: np.ndarray,
        array_metadata: "CoreArrayMetadata",
    ) -> np.ndarray:
        pass

    async def encode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return await _needs_array(chunk_value_handle, array_metadata, self.inner_encode)

    @abstractmethod
    async def inner_encode(
        self,
        chunk_array: np.ndarray,
        array_metadata: "CoreArrayMetadata",
    ) -> np.ndarray:
        pass


class ArrayBytesCodec(Codec):
    input_type = "array"
    output_type = "bytes"

    async def decode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return await _needs_bytes(chunk_value_handle, array_metadata, self.inner_decode)

    @abstractmethod
    async def inner_decode(
        self,
        chunk_array: BytesLike,
        array_metadata: "CoreArrayMetadata",
    ) -> np.ndarray:
        pass

    async def encode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return await _needs_array(chunk_value_handle, array_metadata, self.inner_encode)

    @abstractmethod
    async def inner_encode(
        self,
        chunk_array: np.ndarray,
        array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        pass


class BytesBytesCodec(Codec):
    input_type = "bytes"
    output_type = "bytes"

    async def decode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return await _needs_bytes(chunk_value_handle, array_metadata, self.inner_decode)

    @abstractmethod
    async def inner_decode(
        self,
        chunk_array: BytesLike,
        array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        pass

    async def encode(
        self,
        chunk_value_handle: ValueHandle,
        array_metadata: "CoreArrayMetadata",
    ) -> ValueHandle:
        return await _needs_bytes(chunk_value_handle, array_metadata, self.inner_encode)

    @abstractmethod
    async def inner_encode(
        self,
        chunk_array: BytesLike,
        array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        pass


@frozen
class BloscCodec(BytesBytesCodec):
    configuration: BloscCodecConfigurationMetadata

    @classmethod
    def from_metadata(cls, codec_metadata: BloscCodecMetadata) -> "BloscCodec":
        return cls(
            configuration=codec_metadata.configuration,
        )

    def _get_blosc_codec(self):
        config_dict = asdict(self.configuration)
        map_shuffle_str_to_int = {"noshuffle": 0, "shuffle": 1, "bitshuffle": 2}
        config_dict["shuffle"] = map_shuffle_str_to_int[config_dict["shuffle"]]
        return Blosc.from_config(config_dict)

    async def inner_decode(
        self,
        chunk_bytes: bytes,
        _array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        return await to_thread(self._get_blosc_codec().decode, chunk_bytes)

    async def inner_encode(
        self,
        chunk_bytes: bytes,
        array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        chunk_array = np.frombuffer(chunk_bytes, dtype=array_metadata.dtype)
        return await to_thread(self._get_blosc_codec().encode, chunk_array)


@frozen
class EndianCodec(ArrayBytesCodec):
    configuration: EndianCodecConfigurationMetadata

    @classmethod
    def from_metadata(cls, codec_metadata: EndianCodecMetadata) -> "EndianCodec":
        return cls(
            configuration=codec_metadata.configuration,
        )

    def _get_byteorder(self, array: np.ndarray) -> Literal["big", "little"]:
        if array.dtype.byteorder == "<":
            return "little"
        elif array.dtype.byteorder == ">":
            return "big"
        else:
            import sys

            return sys.byteorder

    async def inner_decode(
        self,
        chunk_bytes: BytesLike,
        array_metadata: "CoreArrayMetadata",
    ) -> np.ndarray:
        if self.configuration.endian == "little":
            prefix = "<"
        else:
            prefix = ">"
        dtype = np.dtype(f"{prefix}{data_type_to_numpy[array_metadata.data_type]}")
        chunk_array = np.frombuffer(chunk_bytes, dtype)
        return chunk_array

    async def inner_encode(
        self,
        chunk_array: np.ndarray,
        _array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        byteorder = self._get_byteorder(chunk_array)
        if self.configuration.endian != byteorder:
            new_dtype = chunk_array.dtype.newbyteorder(self.configuration.endian)
            chunk_array = chunk_array.astype(new_dtype)
        return chunk_array.tobytes()


@frozen
class TransposeCodec(ArrayArrayCodec):
    configuration: TransposeCodecConfigurationMetadata

    @classmethod
    def from_metadata(cls, codec_metadata: TransposeCodecMetadata) -> "TransposeCodec":
        return cls(
            configuration=codec_metadata.configuration,
        )

    async def inner_decode(
        self,
        chunk_array: np.ndarray,
        array_metadata: "CoreArrayMetadata",
    ) -> np.ndarray:
        new_order = self.configuration.order
        chunk_array = chunk_array.view(np.dtype(array_metadata.data_type.value))
        if isinstance(new_order, tuple):
            chunk_array = chunk_array.transpose(new_order)
        elif new_order == "F":
            chunk_array = chunk_array.reshape(
                array_metadata.chunk_shape,
                order="F",
            )
        else:
            chunk_array = chunk_array.reshape(
                array_metadata.chunk_shape,
                order="C",
            )
        return chunk_array

    async def inner_encode(
        self,
        chunk_array: np.ndarray,
        _array_metadata: "CoreArrayMetadata",
    ) -> np.ndarray:
        new_order = self.configuration.order
        if isinstance(new_order, tuple):
            chunk_array = chunk_array.transpose(new_order).reshape(-1, order="C")
        else:
            chunk_array = chunk_array.ravel(order=new_order)
        return chunk_array


@frozen
class GzipCodec(BytesBytesCodec):
    configuration: GzipCodecConfigurationMetadata

    @classmethod
    def from_metadata(cls, codec_metadata: GzipCodecMetadata) -> "GzipCodec":
        return cls(
            configuration=codec_metadata.configuration,
        )

    async def inner_decode(
        self,
        chunk_bytes: bytes,
        _array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        return await to_thread(GZip(self.configuration.level).decode, chunk_bytes)

    async def inner_encode(
        self,
        chunk_bytes: bytes,
        _array_metadata: "CoreArrayMetadata",
    ) -> BytesLike:
        return await to_thread(GZip(self.configuration.level).encode, chunk_bytes)


def blosc_codec(
    cname: Literal["lz4", "lz4hc", "blosclz", "zstd", "snappy", "zlib"] = "zstd",
    clevel: int = 5,
    shuffle: Literal["noshuffle", "shuffle", "bitshuffle"] = "noshuffle",
    blocksize: int = 0,
) -> BloscCodecMetadata:
    return BloscCodecMetadata(
        configuration=BloscCodecConfigurationMetadata(
            cname=cname, clevel=clevel, shuffle=shuffle, blocksize=blocksize
        )
    )


def endian_codec(endian: Literal["big", "little"]) -> EndianCodecMetadata:
    return EndianCodecMetadata(configuration=EndianCodecConfigurationMetadata(endian))


def transpose_codec(
    order: Union[Tuple[int, ...], Literal["C", "F"]]
) -> TransposeCodecMetadata:
    return TransposeCodecMetadata(
        configuration=TransposeCodecConfigurationMetadata(order)
    )


def gzip_codec(level: int = 5) -> GzipCodecMetadata:
    return GzipCodecMetadata(configuration=GzipCodecConfigurationMetadata(level))


def sharding_codec(
    chunk_shape: Tuple[int, ...], codecs: List[CodecMetadata] = []
) -> ShardingCodecMetadata:
    return ShardingCodecMetadata(
        configuration=ShardingCodecConfigurationMetadata(chunk_shape, codecs)
    )



import asyncio
import json
from typing import Any, Dict, List, Literal, Optional, Union

import numcodecs
import numpy as np
from attr import asdict, frozen
from numcodecs.compat import ensure_bytes, ensure_ndarray

from zarrita.common import (
    ZARRAY_JSON,
    ZATTRS_JSON,
    BytesLike,
    ChunkCoords,
    Selection,
    SliceSelection,
    make_cattr,
    to_thread,
)
from zarrita.indexing import BasicIndexer, is_total_slice
from zarrita.metadata import ArrayV2Metadata
from zarrita.store import StoreLike, StorePath, make_store_path
from zarrita.sync import sync
from zarrita.value_handle import (
    BufferValueHandle,
    FileValueHandle,
    NoneValueHandle,
    ValueHandle,
)


@frozen
class _AsyncArrayProxy:
    array: "ArrayV2"

    def __getitem__(self, selection: Selection) -> "_AsyncArraySelectionProxy":
        return _AsyncArraySelectionProxy(self.array, selection)


@frozen
class _AsyncArraySelectionProxy:
    array: "ArrayV2"
    selection: Selection

    async def get(self) -> np.ndarray:
        return await self.array.get_async(self.selection)

    async def set(self, value: np.ndarray):
        return await self.array.set_async(self.selection, value)


@frozen
class ArrayV2:
    metadata: ArrayV2Metadata
    attributes: Optional[Dict[str, Any]]
    store_path: StorePath

    @classmethod
    async def create_async(
        cls,
        store: StoreLike,
        *,
        shape: ChunkCoords,
        dtype: np.dtype,
        chunks: ChunkCoords,
        dimension_separator: Literal[".", "/"] = ".",
        fill_value: Optional[Union[None, int, float]] = None,
        order: Literal["C", "F"] = "C",
        filters: Optional[List[Dict[str, Any]]] = None,
        compressor: Optional[Dict[str, Any]] = None,
        attributes: Optional[Dict[str, Any]] = None,
        exists_ok: bool = False,
    ) -> "ArrayV2":
        store_path = make_store_path(store)
        if not exists_ok:
            assert not await (store_path / ZARRAY_JSON).exists_async()

        metadata = ArrayV2Metadata(
            shape=shape,
            dtype=dtype,
            chunks=chunks,
            order=order,
            dimension_separator=dimension_separator,
            fill_value=0 if fill_value is None else fill_value,
            compressor=numcodecs.get_codec(compressor).get_config()
            if compressor is not None
            else None,
            filters=[numcodecs.get_codec(filter).get_config() for filter in filters]
            if filters is not None
            else None,
        )
        array = cls(
            metadata=metadata,
            store_path=store_path,
            attributes=attributes,
        )
        await array._save_metadata()
        return array

    @classmethod
    def create(
        cls,
        store: StoreLike,
        *,
        shape: ChunkCoords,
        dtype: np.dtype,
        chunks: ChunkCoords,
        dimension_separator: Literal[".", "/"] = ".",
        fill_value: Optional[Union[None, int, float]] = None,
        order: Literal["C", "F"] = "C",
        filters: Optional[List[Dict[str, Any]]] = None,
        compressor: Optional[Dict[str, Any]] = None,
        attributes: Optional[Dict[str, Any]] = None,
        exists_ok: bool = False,
    ) -> "ArrayV2":
        return sync(
            cls.create_async(
                store,
                shape=shape,
                dtype=dtype,
                chunks=chunks,
                order=order,
                dimension_separator=dimension_separator,
                fill_value=0 if fill_value is None else fill_value,
                compressor=compressor,
                filters=filters,
                attributes=attributes,
                exists_ok=exists_ok,
            )
        )

    @classmethod
    async def open_async(
        cls,
        store: StoreLike,
    ) -> "ArrayV2":
        store_path = make_store_path(store)
        zarray_bytes, zattrs_bytes = await asyncio.gather(
            (store_path / ZARRAY_JSON).get_async(),
            (store_path / ZATTRS_JSON).get_async(),
        )
        assert zarray_bytes is not None
        return cls.from_json(
            store_path,
            zarray_json=json.loads(zarray_bytes),
            zattrs_json=json.loads(zattrs_bytes) if zattrs_bytes is not None else None,
        )

    @classmethod
    def open(
        cls,
        store: StoreLike,
    ) -> "ArrayV2":
        return sync(cls.open_async(store))

    @classmethod
    def from_json(
        cls,
        store_path: StorePath,
        zarray_json: Any,
        zattrs_json: Optional[Any],
    ) -> "ArrayV2":
        metadata = make_cattr().structure(zarray_json, ArrayV2Metadata)
        out = cls(
            store_path=store_path,
            metadata=metadata,
            attributes=zattrs_json,
        )
        out._validate_metadata()
        return out

    async def _save_metadata(self) -> None:
        def convert(o):
            if isinstance(o, np.dtype):
                if o.fields is None:
                    return o.str
                else:
                    return o.descr
            raise TypeError

        self._validate_metadata()

        await (self.store_path / ZARRAY_JSON).set_async(
            json.dumps(asdict(self.metadata), default=convert).encode(),
        )
        if self.attributes is not None and len(self.attributes) > 0:
            await (self.store_path / ZATTRS_JSON).set_async(
                json.dumps(self.attributes).encode(),
            )
        else:
            await (self.store_path / ZATTRS_JSON).delete_async()

    def _validate_metadata(self) -> None:
        assert len(self.metadata.shape) == len(
            self.metadata.chunks
        ), "`chunks` and `shape` need to have the same number of dimensions."

    @property
    def ndim(self) -> int:
        return len(self.metadata.shape)

    @property
    def shape(self) -> ChunkCoords:
        return self.metadata.shape

    @property
    def dtype(self) -> np.dtype:
        return self.metadata.dtype

    @property
    def async_(self) -> _AsyncArrayProxy:
        return _AsyncArrayProxy(self)

    def __getitem__(self, selection: Selection):
        return sync(self.get_async(selection))

    async def get_async(self, selection: Selection):
        indexer = BasicIndexer(
            selection,
            shape=self.metadata.shape,
            chunk_shape=self.metadata.chunks,
        )

        # setup output array
        out = np.zeros(
            indexer.shape,
            dtype=self.metadata.dtype,
            order=self.metadata.order,
        )

        # reading chunks and decoding them
        await asyncio.gather(
            *[
                self._read_chunk(chunk_coords, chunk_selection, out_selection, out)
                for chunk_coords, chunk_selection, out_selection in indexer
            ]
        )

        if out.shape:
            return out
        else:
            return out[()]

    async def _read_chunk(
        self,
        chunk_coords: ChunkCoords,
        chunk_selection: SliceSelection,
        out_selection: SliceSelection,
        out: np.ndarray,
    ):
        value_handle = FileValueHandle(
            self.store_path / self._encode_chunk_key(chunk_coords)
        )

        chunk_array = await self._decode_chunk(await value_handle.tobytes())
        if chunk_array is not None:
            tmp = chunk_array[chunk_selection]
            out[out_selection] = tmp
        else:
            out[out_selection] = self.metadata.fill_value

    async def _decode_chunk(
        self, chunk_bytes: Optional[BytesLike]
    ) -> Optional[np.ndarray]:
        if chunk_bytes is None:
            return None

        if self.metadata.compressor is not None:
            compressor = numcodecs.get_codec(self.metadata.compressor)
            chunk_array = ensure_ndarray(
                await to_thread(compressor.decode, chunk_bytes)
            )
        else:
            chunk_array = ensure_ndarray(chunk_bytes)

        # ensure correct dtype
        if str(chunk_array.dtype) != self.metadata.dtype:
            chunk_array = chunk_array.view(self.metadata.dtype)

        # apply filters in reverse order
        if self.metadata.filters is not None:
            for filter_metadata in self.metadata.filters[::-1]:
                filter = numcodecs.get_codec(filter_metadata)
                chunk_array = await to_thread(filter.decode, chunk_array)

        # ensure correct chunk shape
        if chunk_array.shape != self.metadata.chunks:
            chunk_array = chunk_array.reshape(
                self.metadata.chunks,
                order=self.metadata.order,
            )

        return chunk_array

    def __setitem__(self, selection: Selection, value: np.ndarray) -> None:
        sync(self.set_async(selection, value))

    async def set_async(self, selection: Selection, value: np.ndarray) -> None:
        chunk_shape = self.metadata.chunks
        indexer = BasicIndexer(
            selection,
            shape=self.metadata.shape,
            chunk_shape=chunk_shape,
        )

        sel_shape = indexer.shape

        # check value shape
        if np.isscalar(value):
            # setting a scalar value
            pass
        else:
            if not hasattr(value, "shape"):
                value = np.asarray(value, self.metadata.dtype)
            assert value.shape == sel_shape
            if value.dtype != self.metadata.dtype:
                value = value.astype(self.metadata.dtype, order="A")

        # merging with existing data and encoding chunks
        await asyncio.gather(
            *[
                self._write_chunk(
                    value,
                    chunk_shape,
                    chunk_coords,
                    chunk_selection,
                    out_selection,
                )
                for chunk_coords, chunk_selection, out_selection in indexer
            ]
        )

    async def _write_chunk(
        self,
        value: np.ndarray,
        chunk_shape: ChunkCoords,
        chunk_coords: ChunkCoords,
        chunk_selection: SliceSelection,
        out_selection: SliceSelection,
    ):
        value_handle = FileValueHandle(
            self.store_path / self._encode_chunk_key(chunk_coords)
        )

        if is_total_slice(chunk_selection, chunk_shape):
            # write entire chunks
            if np.isscalar(value):
                chunk_array = np.empty(
                    chunk_shape,
                    dtype=self.metadata.dtype,
                    order=self.metadata.order,
                )
                chunk_array.fill(value)
            else:
                chunk_array = value[out_selection]
            await self._write_chunk_to_store(value_handle, chunk_array)

        else:
            # writing partial chunks
            # read chunk first
            tmp = await self._decode_chunk(await value_handle.tobytes())

            # merge new value
            if tmp is None:
                chunk_array = np.empty(
                    chunk_shape,
                    dtype=self.metadata.dtype,
                    order=self.metadata.order,
                )
                chunk_array.fill(self.metadata.fill_value)
            else:
                chunk_array = tmp.copy(
                    order=self.metadata.order,
                )  # make a writable copy
            chunk_array[chunk_selection] = value[out_selection]

            await self._write_chunk_to_store(value_handle, chunk_array)

    async def _write_chunk_to_store(
        self, value_handle: ValueHandle, chunk_array: np.ndarray
    ):
        chunk_value: ValueHandle
        if np.all(chunk_array == self.metadata.fill_value):
            # chunks that only contain fill_value will be removed
            chunk_value = NoneValueHandle()
        else:
            chunk_value = await self._encode_chunk(chunk_array)

        # write out chunk
        await value_handle.set_async(chunk_value)

    async def _encode_chunk(self, chunk_array: np.ndarray):
        chunk_array = chunk_array.ravel(order=self.metadata.order)

        if self.metadata.filters is not None:
            for filter_metadata in self.metadata.filters:
                filter = numcodecs.get_codec(filter_metadata)
                chunk_array = await to_thread(filter.encode, chunk_array)

        if self.metadata.compressor is not None:
            compressor = numcodecs.get_codec(self.metadata.compressor)
            if (
                not chunk_array.flags.c_contiguous
                and not chunk_array.flags.f_contiguous
            ):
                chunk_array = chunk_array.copy(order="A")
            encoded_chunk_bytes = ensure_bytes(
                await to_thread(compressor.encode, chunk_array)
            )
        else:
            encoded_chunk_bytes = ensure_bytes(chunk_array)

        return BufferValueHandle(encoded_chunk_bytes)

    def _encode_chunk_key(self, chunk_coords: ChunkCoords) -> str:
        chunk_identifier = self.metadata.dimension_separator.join(
            map(str, chunk_coords)
        )
        return "0" if chunk_identifier == "" else chunk_identifier

    def __repr__(self):
        return f"<Array_v2 {self.store_path}>"



import asyncio
import json
from typing import Any, Dict, Literal, Optional, Union

from attr import asdict, frozen

from zarrita.array_v2 import ArrayV2
from zarrita.common import ZARRAY_JSON, ZATTRS_JSON, ZGROUP_JSON, make_cattr
from zarrita.store import StoreLike, StorePath, make_store_path
from zarrita.sync import sync


@frozen
class GroupV2Metadata:
    zarr_format: Literal[2] = 2


@frozen
class GroupV2:
    metadata: GroupV2Metadata
    store_path: StorePath
    attributes: Optional[Dict[str, Any]] = None

    @classmethod
    async def create_async(
        cls,
        store: StoreLike,
        *,
        attributes: Optional[Dict[str, Any]] = None,
        exists_ok: bool = False,
    ) -> "GroupV2":
        store_path = make_store_path(store)
        if not exists_ok:
            assert not await (store_path / ZGROUP_JSON).exists_async()
        group = cls(
            metadata=GroupV2Metadata(), attributes=attributes, store_path=store_path
        )
        await group._save_metadata()
        return group

    @classmethod
    def create(
        cls,
        store: StoreLike,
        *,
        attributes: Optional[Dict[str, Any]] = None,
        exists_ok: bool = False,
    ) -> "GroupV2":
        return sync(cls.create_async(store, attributes=attributes, exists_ok=exists_ok))

    @classmethod
    async def open_async(cls, store: StoreLike) -> "GroupV2":
        store_path = make_store_path(store)
        zgroup_bytes = await (store_path / ZGROUP_JSON).get_async()
        assert zgroup_bytes is not None
        zattrs_bytes = await (store_path / ZATTRS_JSON).get_async()
        metadata = json.loads(zgroup_bytes)
        attributes = json.loads(zattrs_bytes) if zattrs_bytes is not None else None

        return cls.from_json(store_path, metadata, attributes)

    @classmethod
    def open(cls, store_path: StorePath) -> "GroupV2":
        return sync(cls.open_async(store_path))

    @classmethod
    def from_json(
        cls,
        store_path: StorePath,
        zarr_json: Any,
        attributes: Optional[Dict[str, Any]] = None,
    ) -> "GroupV2":
        group = cls(
            metadata=make_cattr().structure(zarr_json, GroupV2Metadata),
            store_path=store_path,
            attributes=attributes,
        )
        return group

    @staticmethod
    async def open_or_array(store: StoreLike) -> Union[ArrayV2, "GroupV2"]:
        store_path = make_store_path(store)
        zgroup_bytes, zattrs_bytes = await asyncio.gather(
            (store_path / ZGROUP_JSON).get_async(),
            (store_path / ZATTRS_JSON).get_async(),
        )
        attributes = json.loads(zattrs_bytes) if zattrs_bytes is not None else None
        if zgroup_bytes is not None:
            return GroupV2.from_json(store_path, json.loads(zgroup_bytes), attributes)
        zarray_bytes = await (store_path / ZARRAY_JSON).get_async()
        if zarray_bytes is not None:
            return ArrayV2.from_json(store_path, json.loads(zarray_bytes), attributes)
        raise KeyError

    async def _save_metadata(self) -> None:
        await (self.store_path / ZGROUP_JSON).set_async(
            json.dumps(asdict(self.metadata)).encode(),
        )
        if self.attributes is not None and len(self.attributes) > 0:
            await (self.store_path / ZATTRS_JSON).set_async(
                json.dumps(self.attributes).encode(),
            )
        else:
            await (self.store_path / ZATTRS_JSON).delete_async()

    async def get_async(self, path: str) -> Union[ArrayV2, "GroupV2"]:
        return await self.__class__.open_or_array(self.store_path / path)

    def __getitem__(self, path: str) -> Union[ArrayV2, "GroupV2"]:
        return sync(self.get_async(path))

    async def create_group_async(self, path: str, **kwargs) -> "GroupV2":
        return await self.__class__.create_async(self.store_path / path, **kwargs)

    def create_group(self, path: str, **kwargs) -> "GroupV2":
        return sync(self.create_group_async(path))

    async def create_array_async(self, path: str, **kwargs) -> ArrayV2:
        return await ArrayV2.create_async(self.store_path / path, **kwargs)

    def create_array(self, path: str, **kwargs) -> "ArrayV2":
        return sync(self.create_array_async(path, **kwargs))

    def __repr__(self):
        return f"<Group_v2 {self.store_path}>"



from abc import abstractmethod
from typing import Optional, Tuple

import numpy as np

from zarrita.common import BytesLike
from zarrita.store import StorePath


# ValueHandle abstracts over files, byte buffer, and arrays.
# Makes it easy pass around as references with lazy reading (esp. for files).
# Facilitates implementations of codec which either operate on arrays or bytes.
class ValueHandle:
    @abstractmethod
    async def set_async(
        self,
        value: "ValueHandle",
    ):
        pass

    @abstractmethod
    def __getitem__(self, selection: slice) -> "ValueHandle":
        pass

    @abstractmethod
    async def tobytes(self) -> Optional[bytes]:
        pass

    @abstractmethod
    async def toarray(
        self, shape: Optional[Tuple[int, ...]] = None
    ) -> Optional[np.ndarray]:
        pass


class FileValueHandle(ValueHandle):
    store_path: "StorePath"
    selection: Optional[slice] = None

    def __init__(self, store_path: "StorePath"):
        super().__init__()
        self.store_path = store_path

    def __getitem__(self, selection: slice) -> ValueHandle:
        out = FileValueHandle(self.store_path)
        out.selection = selection
        return out

    async def set_async(
        self,
        value: ValueHandle,
    ):
        buf = await value.tobytes()
        if buf:
            await self.store_path.set_async(buf)
        else:
            await self.store_path.delete_async()

    async def tobytes(self) -> Optional[bytes]:
        if self.selection:
            return await self.store_path.get_async(
                (self.selection.start, self.selection.stop)
            )
        return await self.store_path.get_async()

    async def toarray(
        self, shape: Optional[Tuple[int, ...]] = None
    ) -> Optional[np.ndarray]:
        buf = await self.tobytes()
        if buf is not None:
            return np.frombuffer(buf)
        return None


class BufferValueHandle(ValueHandle):
    buf: BytesLike

    def __init__(self, buf: BytesLike) -> None:
        self.buf = buf

    async def set_async(
        self,
        value: ValueHandle,
    ):
        buf = await value.tobytes()
        assert buf is not None
        self.buf = buf

    def __getitem__(self, selection: slice) -> "ValueHandle":
        return BufferValueHandle(self.buf[selection])

    async def tobytes(self) -> Optional[bytes]:
        return self.buf

    async def toarray(
        self, shape: Optional[Tuple[int, ...]] = None
    ) -> Optional[np.ndarray]:
        return np.frombuffer(self.buf)


class ArrayValueHandle(ValueHandle):
    array: np.ndarray

    def __init__(self, array: np.ndarray) -> None:
        super().__init__()
        self.array = array

    async def set_async(
        self,
        value: ValueHandle,
    ):
        array = await value.toarray()
        assert array is not None
        self.array = array

    def __getitem__(self, selection: slice) -> "ValueHandle":
        return ArrayValueHandle(self.array[selection])

    async def tobytes(self) -> Optional[bytes]:
        array = self.array
        if not array.flags.c_contiguous and not array.flags.f_contiguous:
            array = array.copy(order="A")
        return array.tobytes()

    async def toarray(
        self, shape: Optional[Tuple[int, ...]] = None
    ) -> Optional[np.ndarray]:
        return self.array


class NoneValueHandle(ValueHandle):
    async def set_async(self, value: ValueHandle):
        raise NotImplementedError

    def __getitem__(self, _selection: slice) -> "ValueHandle":
        return self

    async def tobytes(self) -> Optional[bytes]:
        return None

    async def toarray(
        self, shape: Optional[Tuple[int, ...]] = None
    ) -> Optional[np.ndarray]:
        return None
